import typing
from functools import lru_cache

from bakplane.bakplane_pb2 import (
    ResolveResourcePathResponse,
    ResolveResourceIntentResponse,
)
from bakplane.extensions.base import (
    WriteRequest,
    ConnectionStringContext,
    ReadResponse,
    ReadAllRequest,
    ReadRequest,
)
from bakplane.extensions.spark import PluginEntry, WriteResponse
from bakplane.extensions.spark.stores.jdbc import JDBCStoreHelper
from cachetools import cached
from cachetools.keys import hashkey

DRIVER_NAME = "com.amazon.redshift.jdbc42.Driver"


def get_registration():
    return PluginEntry(
        name="Redshift",
        code="redshift",
        description="Redshift store",
        author="Isaac Elbaz",
        read_fn=__read_fn,
        read_all_fn=__read_all_fn,
        write_fn=__write_fn,
        connection_string_url_fn=__connection_string_url,
        driver_name=DRIVER_NAME,
    )


@cached(
    cache={},
    key=lambda user, password, host, port, database, schema_name: hashkey(
        host, database
    ),
)
def get_helper(
    user: str,
    password: str,
    host: str,
    port: int,
    database: str,
    schema_name: str,
):
    return JDBCStoreHelper(
        user=user,
        password=password,
        host=host,
        port=port,
        database=database,
        schema_name=schema_name,
        driver="redshift+psycopg2",
        series_schema="series",
    )


def __connection_string_url(c: ConnectionStringContext) -> str:
    return f'jdbc:redshift://{c.context["host"]}:{c.context["port"]}/{c.context["database"]}'


def __write_fn(
    r: WriteRequest, res: ResolveResourcePathResponse
) -> WriteResponse:
    from bakplane.extensions.spark.stores.jdbc import write_fn

    return write_fn(
        r=r,
        res=res,
        driver=DRIVER_NAME,
        url=__connection_string_url(ConnectionStringContext(res.context)),
    )


def __read_all_fn(
    ctx: typing.Any, r: ReadAllRequest, res: ResolveResourcePathResponse
) -> ReadResponse:
    from bakplane.extensions.spark.stores.jdbc import read_all_fn

    return read_all_fn(
        ctx=ctx,
        r=r,
        res=res,
        url=__connection_string_url(ConnectionStringContext(res.context)),
        driver=DRIVER_NAME,
    )


def __read_fn(
    ctx: typing.Any, r: ReadRequest, res: ResolveResourceIntentResponse
) -> ReadResponse:
    from bakplane.extensions.spark.stores.jdbc import read_fn
    from sqlalchemy.dialects import postgresql

    return read_fn(
        ctx=ctx,
        r=r,
        res=res,
        dialect=postgresql.dialect(),
        url=__connection_string_url(ConnectionStringContext(res.context)),
        driver=DRIVER_NAME,
        store_helper=get_helper(
            user=res.context["username"],
            password=res.context["password"],
            host=res.context["host"],
            port=res.context["port"],
            database=res.context["database"],
            schema_name="temp_data",
        ),
    )



import datetime
import typing
from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum

import pandas as pd
from sqlalchemy import Table, Column
from sqlalchemy.sql import FromClause
from sqlalchemy.sql.elements import BooleanClauseList

from bakplane.bakplane_pb2 import (
    Universe,
    ResolveResourceIntentResponse,
    ResolveResourcePathResponse,
)

date_constraint_t = typing.Union[
    typing.Tuple[datetime.datetime, datetime.datetime],
    typing.Tuple[datetime.datetime, datetime.datetime, str],
    typing.Tuple[datetime.datetime, datetime.datetime, str, str],
    datetime.datetime,
]

DT_COLUMN_NAME = "dt"
DEFAULT_SERIES = [
    "B",
    "D",
    "W",
    "M",
    "SM",
    "BM",
    "MS",
    "SMS",
    "BMS",
    "Q",
    "BQ",
    "QS",
    "Y",
    "BY",
    "YS",
    "BYS",
]
INFINITY_PROXY_DATE = datetime.datetime(year=3000, month=12, day=31)


def flatten_date_constraints(
    c: date_constraint_t,
) -> typing.List[datetime.datetime]:
    if isinstance(c, datetime.datetime):
        return [c, INFINITY_PROXY_DATE]
    elif isinstance(c, tuple):
        if len(c) == 2:
            return [c[0], c[1]]
        elif len(c) == 3 or len(c) == 4:
            results = set()

            ranges = pd.bdate_range(start=c[0], end=c[1], freq=c[2])
            for r in ranges:
                results.add(r.to_pydatetime())
            return list(results)
        else:
            raise RuntimeError(
                f"Invalid dating definition: `{c}` is not valid."
            )


class WriteMode(Enum):
    ERROR_IF_EXISTS = 1
    APPEND = 2
    OVERWRITE = 3
    IGNORE = 4


@dataclass
class ExecutionStatistics:
    elapsed_ms: float


@dataclass
class ConnectionStringContext:
    context: typing.Dict[str, str]


@dataclass
class WriteRequest:
    path: str
    warehouse: str
    df: typing.Any
    mode: WriteMode


@dataclass
class ReadAllRequest:
    path: str
    warehouse: str
    columns: typing.List[str]

    partition_column_name: str = "knowledge_hash"
    partition_num: int = None


@dataclass
class ReadRequestBuildingBlocks:
    asset_table: Table
    mapping_table: Table
    selectable: FromClause
    constraints: typing.List[typing.Any]

    effective_dating_range_semantics: str
    valid_dating_range_semantics: str

    effective_dating_table: Table = None
    valid_dating_table: Table = None

    effective_start_dt: datetime = None
    effective_end_dt: datetime = None

    valid_start_dt: datetime = None
    valid_end_dt: datetime = None

    knowledge_dt: datetime.datetime = None
    query: typing.Any = None
    columns: typing.List[Column] = None
    modifier_context: typing.Dict[str, typing.Any] = None


class DataStoreHelper(ABC):
    @abstractmethod
    def create_temporary_table(
        self,
        prefix: str,
        columns: typing.List[Column],
        rows: typing.List[typing.Any],
    ) -> Table:
        pass

    @abstractmethod
    def get_series_table(
        self,
        frequency: typing.Tuple[datetime.datetime, datetime.datetime, str],
        alias: str = None,
    ) -> Table:
        pass


class Modifier:
    @abstractmethod
    def execute(
        self, r: ReadRequestBuildingBlocks, h: DataStoreHelper
    ) -> ReadRequestBuildingBlocks:
        pass


@dataclass
class ReadRequest:
    path: str
    warehouse: str
    universe: Universe

    pointers: typing.Dict[str, str]

    effective_dating: date_constraint_t = None
    valid_dating: date_constraint_t = None

    columns: typing.List[str] = None
    knowledge_dt: datetime.datetime = None

    query_builder: typing.Callable[
        [ReadRequestBuildingBlocks], typing.Any
    ] = None

    additional_constraints: typing.Callable[
        [ReadRequestBuildingBlocks], BooleanClauseList
    ] = None

    effective_dating_hint: typing.Union[
        str, typing.Tuple[datetime.datetime, datetime.datetime]
    ] = None

    modifiers: typing.List[Modifier] = None

    partition_column_name: str = "knowledge_hash"
    partition_num: int = None

    def get_derived_effective_start_dt(self) -> datetime.datetime:
        if self.effective_dating is not None:
            return self.get_min_effective_dt()
        return self.get_min_valid_dt()

    def get_derived_effective_end_dt(self) -> datetime.datetime:
        if self.effective_dating is not None:
            return self.get_max_effective_dt()
        return self.get_max_valid_dt()

    def get_resolved_effective_dating(self) -> typing.List[datetime.datetime]:
        return flatten_date_constraints(self.effective_dating)

    def get_resolved_valid_dating(self) -> typing.List[datetime.datetime]:
        return flatten_date_constraints(self.valid_dating)

    def get_min_effective_dt(self) -> datetime.datetime:
        return min(self.get_resolved_effective_dating())

    def get_max_effective_dt(self) -> datetime.datetime:
        return max(self.get_resolved_effective_dating())

    def get_min_valid_dt(self) -> datetime.datetime:
        return min(self.get_resolved_valid_dating())

    def get_max_valid_dt(self) -> datetime.datetime:
        return max(self.get_resolved_valid_dating())


@dataclass
class ReadUniverseRequest:
    universe: Universe
    effective_start_dt: datetime.datetime
    effective_end_dt: datetime.datetime
    knowledge_dt: datetime.datetime = None


@dataclass
class ReadResponse:
    df: typing.Any
    execution_statistics: ExecutionStatistics
    context: typing.Dict[str, str]


@dataclass
class WriteResponse:
    execution_statistics: ExecutionStatistics


@dataclass
class ReadUniverseResponse:
    df: typing.Any
    execution_statistics: ExecutionStatistics


@dataclass
class PluginEntry:
    name: str
    code: str
    description: str
    author: str
    write_fn: typing.Callable[
        [WriteRequest, ResolveResourcePathResponse], WriteResponse
    ]
    read_fn: typing.Callable[
        [typing.Any, ReadRequest, ResolveResourceIntentResponse], ReadResponse
    ]
    read_all_fn: typing.Callable[
        [typing.Any, ReadAllRequest, ResolveResourcePathResponse], ReadResponse
    ]
    connection_string_url_fn: typing.Callable[[ConnectionStringContext], str]
    driver_name: str


class BaseExtension(ABC):
    @abstractmethod
    def read(self, r: ReadRequest) -> ReadResponse:
        pass

    @abstractmethod
    def read_all(self, r: ReadAllRequest) -> ReadResponse:
        pass

    @abstractmethod
    def write(self, r: WriteRequest) -> WriteResponse:
        pass

    @abstractmethod
    def read_universe(self, r: ReadUniverseRequest) -> ReadUniverseResponse:
        pass



#!/usr/bin/env python

import json
import os
import pathlib
import typing
from abc import ABC, abstractmethod
from functools import lru_cache

import boto3
import click
import sqlalchemy.types as types
from sqlalchemy import PrimaryKeyConstraint
from sqlalchemy import (
    Table,
    Column,
    String,
    MetaData,
    TIMESTAMP,
    BigInteger,
    Index,
)
from sqlalchemy.dialects import postgresql
from sqlalchemy.dialects.postgresql.base import BYTEA
from sqlalchemy.schema import CreateTable
from sqlalchemy.sql.ddl import CreateIndex, DropTable
from sqlalchemy.sql.sqltypes import (
    BOOLEAN,
    DATE,
    DATETIME,
    DECIMAL,
    FLOAT,
    Integer,
    JSON,
    TIME,
    TEXT,
)
from sqlalchemy_redshift.dialect import RedshiftDialect

UP_TABLES_MIGRATION_NAME = "1_tables.up.sql"
DOWN_TABLES_MIGRATION_NAME = "1_tables.down.sql"


class VarcharMax(types.UserDefinedType):
    def get_col_spec(self):
        return "VARCHAR(MAX)"


class MigrationProvider(ABC):
    def create_table(
        self,
        name: str,
        schema: str,
        columns: typing.List[Column],
        primary_key_columns: typing.List[str],
        **kwargs,
    ) -> Table:
        return Table(
            name,
            self.get_metadata(),
            *columns,
            PrimaryKeyConstraint(*primary_key_columns),
            schema=schema,
            **kwargs,
        )

    def get_dating_columns(self):
        return [
            "effective_start_dt",
            "effective_end_dt",
            "valid_start_dt",
            "valid_end_dt",
            "knowledge_start_dt",
            "knowledge_end_dt",
        ]

    def get_hashing_columns(self):
        return ["knowledge_hash", "payload_hash"]

    def supports_partitioning(self):
        return False

    def supports_indices(self):
        return False

    def get_partition_ranges(self):
        return [1900] + list(range(1995, 2030)) + [3000]

    def get_metadata(self) -> MetaData:
        return MetaData()

    @abstractmethod
    def get_sqlalchemy_dialect(self):
        pass

    @abstractmethod
    def get_name(self):
        pass

    def add_partitioning_constraint(self, tbl: Table):
        pass

    def get_auxiliary_columns(self):
        return [
            Column("ingestion_session_id", BigInteger, nullable=False),
            Column("resource_code", String(128), nullable=False),
        ]

    def create_partition_table(
        self, f, schema: str, parent_name: str, _from: str, to: str
    ):
        pass

    def drop_partition_table(
        self, f, schema: str, parent_name: str, _from: str, to: str
    ):
        pass

    def map_attribute_type(self, t: str, col):
        pass


class PostgresProvider(MigrationProvider):
    def supports_indices(self):
        return True

    def supports_partitioning(self):
        return True

    def get_sqlalchemy_dialect(self):
        return postgresql.dialect()

    def get_name(self):
        return "postgres"

    def drop_partition_table(
        self, f, schema: str, parent_name: str, _from: str, to: str
    ):
        stmt = f"DROP TABLE {schema}.{parent_name}_p{_from}_{to};\n"
        f.write(stmt)

    def create_partition_table(
        self, f, schema: str, parent_name: str, _from: str, to: str
    ):
        stmt = f"CREATE TABLE {schema}.{parent_name}_p{_from}_{to} PARTITION OF {schema}.{parent_name} FOR VALUES FROM ('{_from}-01-01') TO ('{to}-01-01');\n"
        f.write(stmt)

    def add_partitioning_constraint(
        self, tbl: Table, column: str = "effective_start_dt"
    ):
        tbl.dialect_options["postgresql"]["partition_by"] = f"RANGE({column})"

    def map_attribute_type(self, t: str, col):
        return map_attribute_type(t, col)


class RedshiftProvider(MigrationProvider):
    def create_table(
        self,
        name: str,
        schema: str,
        columns: typing.List[Column],
        primary_key_columns: typing.List[str],
        **kwargs,
    ) -> Table:
        return super(RedshiftProvider, self).create_table(
            name, schema, columns, primary_key_columns
        )

    def supports_indices(self):
        return False

    def supports_partitioning(self):
        return False

    def get_sqlalchemy_dialect(self):
        return RedshiftDialect()

    def get_name(self):
        return "redshift"

    def map_attribute_type(self, t: str, col):
        if t == "text":
            return VarcharMax
        else:
            return map_attribute_type(t, col)


class MigrationProviderFactory(object):
    @staticmethod
    @lru_cache()
    def get(warehouse):
        if warehouse == "postgres" or warehouse == "postgresql":
            return PostgresProvider()
        elif warehouse == "redshift":
            return RedshiftProvider()
        raise ValueError("Invalid warehouse provided.")


def create_schema_if_needed(
    f_up,
    f_down,
    provider: MigrationProvider,
    doc: typing.Dict[str, typing.Any],
):
    required_schemas = [doc["provider"]]

    for r in doc["resources"]:
        required_schemas.append(r["dataset"])

    for rs in set(required_schemas):
        f_up.write(f"CREATE SCHEMA {rs};\n")
        f_down.write(f"DROP SCHEMA {rs};\n")


def create_table_if_needed(
    f_up,
    f_down,
    provider: MigrationProvider,
    r: typing.Dict[str, typing.Any],
    schema: str,
):
    columns = []
    indices = [
        ["knowledge_hash", "payload_hash", "resource_code"],
    ]

    ix_count = 0

    found_context = False
    for ctx in r["context"]:
        if ctx["warehouse_kind"] == provider.get_name():
            found_context = True
            for p in ctx["pointers"]:
                columns.append(
                    Column(p["column_name"], BigInteger, nullable=False)
                )

                indices.append(
                    [
                        p["column_name"],
                        "effective_start_dt",
                        "effective_end_dt",
                    ]
                )

    if not found_context:
        if "pointers" in r and len(r["pointers"]) > 0:
            for p in r["pointers"]:
                columns.append(
                    Column(
                        f'{p["entity_type"]}_identity',
                        BigInteger,
                        nullable=False,
                    )
                )

    # add attributes
    for c in r["schema"]["attributes"]:
        columns.append(
            Column(
                c["name"],
                provider.map_attribute_type(c["data_type"], c),
                nullable="mode" in c and c["mode"] == "nullable",
            )
        )

    # add dating columns
    for dc in provider.get_dating_columns():
        columns.append(Column(dc, TIMESTAMP, nullable=False))

    # add hashing
    for hc in provider.get_hashing_columns():
        columns.append(Column(hc, BigInteger, nullable=False))

    columns.extend(provider.get_auxiliary_columns())

    tbl = provider.create_table(
        r["code"],
        schema,
        columns,
        [
            "knowledge_hash",
            "payload_hash",
            "knowledge_end_dt",
            "effective_start_dt",
            "valid_start_dt",
        ],
    )

    if provider.supports_partitioning():
        provider.add_partitioning_constraint(tbl)

    tbl_up_statement = str(
        CreateTable(tbl).compile(dialect=provider.get_sqlalchemy_dialect())
    ).rstrip()

    tbl_down_statement = str(
        DropTable(tbl).compile(dialect=provider.get_sqlalchemy_dialect())
    ).rstrip()

    f_up.write(f"{tbl_up_statement};\n")
    f_down.write(f"{tbl_down_statement};\n")

    if provider.supports_indices():
        for i in indices:
            ix = Index(
                f'{r["code"]}_ix_{ix_count}', *[tbl.c.get(x) for x in i]
            )
            ix_count += 1

            cx = str(
                CreateIndex(ix).compile(
                    dialect=provider.get_sqlalchemy_dialect()
                )
            )

            f_up.write(f"{cx};\n")

    if provider.supports_partitioning():
        ranges = provider.get_partition_ranges()
        for i in range(0, len(ranges) - 1):
            provider.create_partition_table(
                f_up, schema, r["code"], str(ranges[i]), str(ranges[i + 1])
            )

    return tbl


def map_attribute_type(t: str, col):
    t = t.lower()
    if t == "string":
        return String(col.get("length", 128))
    if t == "text":
        return TEXT
    elif t == "integer":
        return Integer
    elif t == "numeric":
        scale = col.get("scale", 4)
        precision = col.get("precision", 15)
        if scale and precision and scale > 0 and precision > 0:
            return DECIMAL(scale=scale, precision=precision)
        else:
            return DECIMAL()
    elif t == "timestamp":
        return TIMESTAMP
    elif t == "boolean":
        return BOOLEAN
    elif t == "floating":
        return FLOAT
    elif t == "bytes":
        return BYTEA
    elif t == "date":
        return DATE
    elif t == "datetime":
        return DATETIME
    elif t == "time":
        return TIME
    elif t == "json":
        return JSON
    elif t == "bigint":
        return BigInteger

    raise ValueError(f"Invalid field type `{t}`.")


def handle_adapter_migrations():
    return None, None


def handle_resource_migrations(
    provider: MigrationProvider,
    r: typing.Dict[str, typing.Any],
    doc: typing.Dict[str, typing.Any],
):
    up_file = os.path.join(
        "migrations", r["code"], provider.get_name(), UP_TABLES_MIGRATION_NAME
    )

    down_file = os.path.join(
        "migrations",
        r["code"],
        provider.get_name(),
        DOWN_TABLES_MIGRATION_NAME,
    )

    pathlib.Path(
        os.path.join("migrations", r["code"], provider.get_name())
    ).mkdir(parents=True, exist_ok=True)

    click.echo(f"Adding table migrations to `{up_file}` ...")

    with open(up_file, "w") as f_up:
        with open(down_file, "w") as f_down:
            create_table_if_needed(f_up, f_down, provider, r, doc["provider"])

    click.echo(f"Table up migrations successfully written to `{up_file}`.")
    click.echo(
        f"Table down  migrations successfully written to `{down_file}`."
    )

    return up_file, down_file


def generate_initial_migration_for_resource(
    provider: MigrationProvider,
    r: typing.Dict[str, typing.Any],
    doc: typing.Dict[str, typing.Any],
    bucket: str,
    upload: bool,
):
    client = boto3.client("s3")

    (
        tbl_migration_file_up,
        tbl_migration_file_down,
    ) = handle_resource_migrations(provider, r, doc)

    tbl_migration_dst_path = (
        doc["code"] + "/" + r["code"] + "/" + provider.get_name() + "/"
    )

    if upload:
        click.echo(f"Uploading table migration to `{tbl_migration_dst_path}.")

        client.upload_file(
            tbl_migration_file_up,
            bucket,
            tbl_migration_dst_path + os.path.basename(tbl_migration_file_up),
        )
        client.upload_file(
            tbl_migration_file_down,
            bucket,
            tbl_migration_dst_path + os.path.basename(tbl_migration_file_down),
        )


def generate_initial_migrations(
    provider: MigrationProvider,
    doc: typing.Dict[str, typing.Any],
    bucket: str,
    upload: bool,
):
    for r in doc["resources"]:
        generate_initial_migration_for_resource(
            provider, r, doc, bucket, upload
        )


@click.command()
@click.option(
    "--metadata", "-m", required=True, help="The metadata resource path."
)
@click.option(
    "--warehouse",
    "-w",
    multiple=True,
    required=True,
    help="You can provide multiple warehouse destinations.",
)
@click.option(
    "--bucket",
    "-b",
    default="bakplane-dev-migrations-internal",
    required=True,
    help="The bucket containing the migrations.",
)
@click.option(
    "--upload",
    "-u",
    default=True,
    required=True,
    help="The bucket containing the migrations.",
)
def main(metadata, warehouse, bucket, upload):
    with open(metadata, "r") as f:
        doc = json.load(f)
        for w in warehouse:
            generate_initial_migrations(
                MigrationProviderFactory.get(w), doc, bucket, upload
            )


if __name__ == "__main__":
    main()



import ctypes
import farmhash
import json
import typing
from datetime import datetime
from timeit import default_timer as timer

from google.protobuf.timestamp_pb2 import Timestamp

from bakplane.bakplane_pb2 import TimestampRange
from bakplane.extensions.base import ExecutionStatistics


def hash_string(s: str) -> int:
    return ctypes.c_long(farmhash.fingerprint64(s)).value


def to_timestamp(dt: datetime) -> int:
    ts = Timestamp()
    ts.FromDatetime(dt)

    return ts.seconds


def to_proto_timestamp(dt: datetime) -> Timestamp:
    ts = Timestamp()
    ts.FromDatetime(dt)

    return ts


def to_proto_timestamp_range(s: datetime, e: datetime) -> TimestampRange:
    return TimestampRange(
        start_dt=to_proto_timestamp(s), end_dt=to_proto_timestamp(e)
    )


def to_execution_statistics(s: timer, e: timer) -> ExecutionStatistics:
    return ExecutionStatistics(e - s)


def parse_value_by_type(s: str, t: int) -> typing.Any:
    if t is None:
        return None

    if t == 0:
        return str(s)
    elif t == 1:
        return int(s)
    elif t == 2:
        return int(s)
    elif t == 3:
        return datetime.strptime(s, "%Y-%m-%d %H:%M:%S.%f")
    elif t == 4:
        return s == "true" or s == "TRUE" or s == "1" or s == 1
    elif t == 5:
        return float(s)
    elif t == 6:
        return s
    elif t == 7:
        return datetime.strptime(s, "%Y-%m-%d")
    elif t == 8:
        return datetime.strptime(s, "%Y-%m-%d %H:%M:%S.%f")
    elif t == 9:
        return datetime.strptime(s, "%H:%M:%S.%f")
    elif t == 10:
        return json.loads(s)


def ts(d: str, f: str = "%Y-%m-%d") -> Timestamp:
    t = Timestamp()
    t.FromDatetime(datetime.strptime(d, f))
    return t


def tsr(s: str, e: str, f: str = "%Y-%m-%d") -> TimestampRange:
    return TimestampRange(start_dt=ts(s, f), end_dt=ts(e, f))



import datetime
import os
import tempfile
import typing
import logging
import json

from dataclasses import dataclass

from dominus.dominus_pb2 import (
    EntityEntry,
    Descriptor as PDescriptor,
    IdentityEntry,
    TimestampRange,
    RelationshipEntry,
)

from bakplane.bakplane_pb2 import ErrorEntry
from bakplane.utils import to_proto_timestamp, to_timestamp, hash_string


@dataclass
class IngestionSessionContext:
    ingestion_session_id: int
    entities_output_location: str
    assets_output_location: str
    errors_output_location: str
    relationships_output_location: str
    debug_metadata_path: str = None
    debug_asset_to_attributes: typing.Dict[str, typing.Set[str]] = None

    def __post_init__(self):
        if self.debug_metadata_path is not None:
            if os.path.exists(self.debug_metadata_path):
                with open(self.debug_metadata_path, "r") as f:
                    local_metadata = json.load(f)
                    if "resources" in local_metadata:
                        self.debug_asset_to_attributes = {}
                        for r in local_metadata["resources"]:
                            if "schema" in r and "attributes" in r["schema"]:
                                for attr in r["schema"]["attributes"]:
                                    if (
                                        r["code"]
                                        not in self.debug_asset_to_attributes
                                    ):
                                        self.debug_asset_to_attributes[
                                            r["code"]
                                        ] = set()
                                    self.debug_asset_to_attributes[
                                        r["code"]
                                    ].add(attr["name"])
                logging.info(
                    f"Extracted the following assets to attributes: `{self.debug_asset_to_attributes}."
                )

    def validate_asset(self, asset):
        if (
            self.debug_metadata_path is None
            or asset.resource_code not in self.debug_asset_to_attributes
        ):
            return True

        expected_fields = self.debug_asset_to_attributes[asset.resource_code]
        found_fields = set(asset.fields.keys())

        missing_fields = expected_fields - found_fields
        if len(missing_fields) > 0:
            raise RuntimeError(
                f"Asset `{asset}`` is missing the following fields: `{missing_fields}`."
            )

        extraneous_fields = found_fields - expected_fields
        if len(extraneous_fields) > 0:
            raise RuntimeError(
                f"Asset `{asset}` contain unknown fields: `{extraneous_fields}`."
            )

    @staticmethod
    def from_env():
        return IngestionSessionContext(
            ingestion_session_id=int(os.getenv("INGESTION_SESSION_ID", 0)),
            entities_output_location=os.getenv(
                "ENTITIES_OUTPUT_LOCATION", f"file://{tempfile.mkdtemp()}/"
            ),
            assets_output_location=os.getenv(
                "ASSETS_OUTPUT_LOCATION", f"file://{tempfile.mkdtemp()}/"
            ),
            errors_output_location=os.getenv(
                "ERRORS_OUTPUT_LOCATION", f"file://{tempfile.mkdtemp()}/"
            ),
            relationships_output_location=os.getenv(
                "RELATIONSHIPS_OUTPUT_LOCATION",
                f"file://{tempfile.mkdtemp()}/",
            ),
            debug_metadata_path=os.getenv("DEBUG_METADATA_PATH", None),
        )


@dataclass
class Asset:
    resource_code: str
    retrieved_by: typing.Dict[str, int]
    fields: typing.Dict[str, str]

    key_hash_fields: typing.List[str]
    payload_hash_fields: typing.List[str]

    effective_start_dt: datetime.datetime
    effective_end_dt: datetime.datetime
    valid_start_dt: datetime.datetime = None
    valid_end_dt: datetime.datetime = None
    knowledge_start_dt: datetime.datetime = None
    knowledge_end_dt: datetime.datetime = None

    def calculate_hash(self, f: typing.List[str]):
        parts = set()

        for k in f:
            if k in self.fields:
                parts.add(str(self.fields[k]))
            elif k in self.retrieved_by:
                parts.add(str(self.retrieved_by[k]))
            elif k == "effective_start_dt":
                parts.add(str(to_timestamp(self.effective_start_dt)))
            elif k == "effective_end_dt":
                parts.add(str(to_timestamp(self.effective_end_dt)))
            elif k == "valid_start_dt":
                if self.valid_start_dt:
                    parts.add(str(to_timestamp(self.valid_start_dt)))
            elif k == "valid_end_dt":
                if self.valid_end_dt:
                    parts.add(str(to_timestamp(self.valid_end_dt)))
            elif k == "knowledge_start_dt":
                if self.knowledge_start_dt:
                    parts.add(str(to_timestamp(self.knowledge_start_dt)))
            elif k == "knowledge_end_dt":
                if self.knowledge_end_dt:
                    parts.add(str(to_timestamp(self.knowledge_end_dt)))

        return hash_string("".join(sorted(list(parts))))

    @property
    def key_hash(self):
        return self.calculate_hash(self.key_hash_fields)

    @property
    def payload_hash(self):
        return self.calculate_hash(self.payload_hash_fields)

    def to_csv_entry(
        self, ctx: IngestionSessionContext, csv_separator: str, null_value: str
    ):
        parts = []

        sorted_retrieved_by = sorted(self.retrieved_by.keys())
        sorted_fields = sorted(self.fields.keys())

        for k in sorted_retrieved_by:
            parts.append(str(self.retrieved_by[k]))

        for k in sorted_fields:
            parts.append(str(self.fields[k]))

        parts.append(
            self.effective_start_dt.strftime("%Y-%m-%d") + " 00:00:00.00000"
        )
        parts.append(
            self.effective_end_dt.strftime("%Y-%m-%d") + " 00:00:00.00000"
        )

        if self.valid_start_dt is None:
            self.valid_start_dt = self.effective_end_dt

        if self.valid_end_dt is None:
            self.valid_end_dt = datetime.datetime(year=3000, month=12, day=31)

        if self.knowledge_start_dt is None:
            self.knowledge_start_dt = datetime.datetime.utcnow()

        if self.knowledge_end_dt is None:
            self.knowledge_end_dt = datetime.datetime(
                year=3000, month=12, day=31
            )

        parts.append(
            self.valid_start_dt.strftime("%Y-%m-%d") + " 00:00:00.00000"
        )
        parts.append(
            self.valid_end_dt.strftime("%Y-%m-%d") + " 00:00:00.00000"
        )

        parts.append(
            self.knowledge_start_dt.strftime("%Y-%m-%d") + " 00:00:00.00000"
        )
        parts.append(
            self.knowledge_end_dt.strftime("%Y-%m-%d") + " 00:00:00.00000"
        )

        parts.append(self.resource_code)
        parts.append(str(ctx.ingestion_session_id))

        parts.append(str(self.key_hash))
        parts.append(str(self.payload_hash))

        return csv_separator.join(
            [
                x.strip()
                .replace(csv_separator, "")
                .replace("\n", "")
                .replace("\r", "")
                if x is not None
                else null_value
                for x in parts
            ]
        )


@dataclass
class Identity:
    entity_type: str
    perspective: str
    attributes: typing.Dict[str, str]

    @property
    def uid(self):
        c = []
        for i in sorted(self.attributes.keys()):
            c.append(i + self.attributes[i])

        return hash_string(
            "".join([self.perspective, self.entity_type, "".join(c)])
        )

    def to_proto(self) -> IdentityEntry:
        return IdentityEntry(
            uid=self.uid,
            entity_type_code=self.entity_type,
            perspective_code=self.perspective,
        )


@dataclass
class Descriptor:
    attribute: str
    value: str
    effective_dating: typing.List[
        typing.Tuple[datetime.datetime, datetime.datetime]
    ]
    knowledge_dating: typing.List[
        typing.Tuple[datetime.datetime, datetime.datetime]
    ] = None

    def to_proto(self) -> PDescriptor:
        return PDescriptor(
            attribute_code=self.attribute,
            value=self.value,
            effective_dating=[
                TimestampRange(
                    start_dt=to_proto_timestamp(x[0]),
                    end_dt=to_proto_timestamp(x[1]),
                )
                for x in self.effective_dating
            ],
        )


@dataclass
class Entity:
    identity: Identity
    descriptors: typing.List[Descriptor]

    def to_proto(self) -> EntityEntry:
        return EntityEntry(
            identity=self.identity.to_proto(),
            descriptors=[x.to_proto() for x in self.descriptors],
        )


@dataclass
class Relationship:
    source_identity_uid: int
    target_identity_uid: int
    relationship_type: str
    effective_dating: typing.List[
        typing.Tuple[datetime.datetime, datetime.datetime]
    ]

    @property
    def identity(self):
        return (
            hash(self.source_identity_uid)
            + hash(self.target_identity_uid)
            + hash(self.relationship_type)
        )

    def to_proto(self) -> RelationshipEntry:
        return RelationshipEntry(
            source_identity_uid=self.source_identity_uid,
            target_identity_uid=self.target_identity_uid,
            relationship_type_code=self.relationship_type,
            effective_dating=[
                TimestampRange(
                    start_dt=to_proto_timestamp(x[0]),
                    end_dt=to_proto_timestamp(x[1]),
                )
                for x in self.effective_dating
            ],
        )


@dataclass
class Error:
    message: str
    stacktrace: str = None
    context: typing.Dict[str, str] = None

    def to_proto(self) -> ErrorEntry:
        return ErrorEntry(
            message=self.message,
            stacktrace=self.stacktrace,
            context=self.context,
        )



from bakplane.extensions.base import WriteMode
from datetime import datetime, timedelta


def to_spark_save_mode(mode: WriteMode) -> str:
    if mode == WriteMode.APPEND:
        return "append"

    if mode == WriteMode.ERROR_IF_EXISTS:
        return "errorifexists"

    if mode == WriteMode.IGNORE:
        return "ignore"

    return "overwrite"


def timestamp_to_datetime(timestamp: int):
    if timestamp < 0:
        return datetime(1970, 1, 1) + timedelta(seconds=timestamp)
    else:
        return datetime.utcfromtimestamp(timestamp)


def timestamp_to_date_string(timestamp: int, f="%Y-%m-%d"):
    dt = timestamp_to_datetime(timestamp)
    return dt.strftime(f)


def datetime_to_date_string(dt: datetime, f="%Y-%m-%d"):
    return dt.strftime(f)



import logging
import gzip
import os
import shutil
import tempfile
import glob

from abc import ABC, abstractmethod
from urllib.parse import urlparse

import boto3
import typing


class LandingZone(ABC):
    @abstractmethod
    def upload(
        self,
        src_path: str,
        dst_path: str,
        suffix: str = "",
        prefix: str = None,
    ):
        pass

    @abstractmethod
    def list_files(self, bucket: str, prefix: str) -> typing.List[str]:
        pass

    @abstractmethod
    def download_file_into(
        self, bucket: str, prefix: str, f
    ) -> typing.List[str]:
        pass


class FilesystemLandingZone(LandingZone):
    def list_files(self, bucket: str, prefix: str) -> typing.List[str]:
        return glob.glob(os.path.join(bucket, prefix), recursive=True)

    def download_file_into(
        self, bucket: str, prefix: str, f
    ) -> typing.List[str]:
        with open(os.path.join(bucket, prefix), "r+b") as f_in:
            shutil.copyfileobj(f_in, f)

    def upload(
        self,
        src_path: str,
        dst_path: str,
        suffix: str = "",
        prefix: str = None,
    ):
        u = urlparse(dst_path)

        p = u.path
        if prefix:
            p = os.path.join(p, prefix) + "/"

        if not os.path.exists(p):
            os.makedirs(p)

        with open(src_path, "rb") as f_in:
            with tempfile.NamedTemporaryFile(
                "wb",
                prefix=p,
                delete=False,
                suffix=suffix,
            ) as f:
                with gzip.open(f, "wb") as f_out:
                    shutil.copyfileobj(f_in, f_out)
                    logging.info(f"Artifact written to `{f.name}`")


class AmazonS3LandingZone(LandingZone):
    def __init__(self):
        self.client = boto3.client("s3")

    def list_files(self, bucket: str, prefix: str) -> typing.List[str]:
        kwargs = {"Bucket": bucket, "Prefix": prefix}
        while True:
            resp = self.client.list_objects_v2(**kwargs)
            for obj in resp["Contents"]:
                yield obj["Key"]
            try:
                kwargs["ContinuationToken"] = resp["NextContinuationToken"]
            except KeyError:
                break

    def download_file_into(self, bucket: str, key: str, f) -> typing.List[str]:
        return self.client.download_fileobj(bucket, key, f)

    def upload(
        self,
        src_path: str,
        dst_path: str,
        suffix: str = "",
        prefix: str = None,
    ):
        u = urlparse(dst_path)

        if u.scheme != "s3":
            raise RuntimeError("Invalid destination path.")

        with open(src_path, "rb") as f_in:
            f = tempfile.NamedTemporaryFile(
                delete=False, prefix=os.path.basename(src_path)
            )
            final_path = f.name

            with gzip.open(f, "wb") as f_out:
                shutil.copyfileobj(f_in, f_out)

            f.close()

            self.client.upload_file(
                final_path,
                u.netloc,
                u.path[1:] + os.path.basename(src_path) + suffix,
            )


class LandingZoneFactory(object):
    @staticmethod
    def build_from_code(code):
        if code == "s3":
            return AmazonS3LandingZone()
        elif code == "file":
            return FilesystemLandingZone()
        raise RuntimeError(
            f"Could not find landing zone matching scheme `{u.scheme}`."
        )

    @staticmethod
    def build_from_uri(uri):
        u = urlparse(uri)

        if u.scheme == "s3":
            return AmazonS3LandingZone()
        elif u.scheme == "file":
            return FilesystemLandingZone()
        raise RuntimeError(
            f"Could not find landing zone matching scheme `{u.scheme}`."
        )
